/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lislib.h"
#include "lis_matrix.h"

int lis_matrix_copy(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int err;

	if( Ain==NULL )
	{
		SETERR("matrix Ain is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type==LIS_MATRIX_NULL )
	{
		SETERR("matrix Ain is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type<=LIS_MATRIX_ASSEMBLING )
	{
		SETERR("matrix Ain is assembling\n");
		return LIS_FAILS;
	}

	switch( Ain->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_matrix_copy_crs(Ain,Aout);
		break;
	case LIS_MATRIX_CCS:
		err = lis_matrix_copy_ccs(Ain,Aout);
		break;
	case LIS_MATRIX_MSR:
		err = lis_matrix_copy_msr(Ain,Aout);
		break;
	case LIS_MATRIX_DIA:
		err = lis_matrix_copy_dia(Ain,Aout);
		break;
	case LIS_MATRIX_ELL:
		err = lis_matrix_copy_ell(Ain,Aout);
		break;
	case LIS_MATRIX_JDS:
		err = lis_matrix_copy_jds(Ain,Aout);
		break;
	case LIS_MATRIX_BSR:
		err = lis_matrix_copy_bsr(Ain,Aout);
		break;
	case LIS_MATRIX_BSC:
		err = lis_matrix_copy_bsc(Ain,Aout);
		break;
	case LIS_MATRIX_VBR:
		err = lis_matrix_copy_vbr(Ain,Aout);
		break;
	case LIS_MATRIX_DNS:
		err = lis_matrix_copy_dns(Ain,Aout);
		break;
	case LIS_MATRIX_COO:
		err = lis_matrix_copy_coo(Ain,Aout);
		break;
	}

	if( err )
	{
		if( err==LIS_NOT_IMPLEMENTED )
		{
			SETERR("not implemented\n");
		}
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
}

int lis_matrix_copy_crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
			g_index[j] = Ain->g_index[j];
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_ccs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
			g_index[j] = Ain->g_index[j];
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}
}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = Ain->ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_msr(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	ndz     = Ain->ndz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	g_index = NULL;

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy msr */
	for(i=0;i<n+1;i++)
	{
		value[i]   = Ain->value[i];
		index[i]   = Ain->index[i];
		g_index[i] = Ain->g_index[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
			g_index[j] = Ain->g_index[j];
		}
	}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	ndz     = Ain->ndz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy msr */
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		value[i]   = Ain->value[i];
		index[i]   = Ain->index[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}
}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	ndz     = Ain->ndz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy msr */
	for(i=0;i<n+1;i++)
	{
		value[i]   = Ain->value[i];
		index[i]   = Ain->index[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->index[j];
		}
	}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_dia(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnd,is,nprocs,my_rank;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnd     = Ain->nnd;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	g_index = NULL;

	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( gn*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dia */
	for(i=0;i<nnd;i++)
	{
		index[i]   = Ain->index[i];
	}
	for(i=0;i<gn;i++)
	{
		g_index[i] = Ain->g_index[i];
	}
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnd,is,ie,nprocs,my_rank;
	int			*index,*col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	nnd     = Ain->nnd;
	is      = Ain->is;
	nprocs  = omp_get_max_threads();

	index   = NULL;
	value   = NULL;
	col     = NULL;

	col = (int *)lis_malloc( nprocs*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(3,index,value,col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nprocs*nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,index,value,col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,value,col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,value,col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dia */
#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	k       = ie-is;

	col[my_rank] = Ain->col[my_rank];
	for(j=0;j<nnd;j++)
	{
		index[my_rank*nnd + j] = Ain->index[my_rank*nnd + j];
	}
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<ie-is;i++)
		{
			value[is*nnd + j*k+i]   = Ain->value[is*nnd + j*k+i];
		}
	}
}

	err = lis_matrix_set_dia(col,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,value,col);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnd,is,nprocs,my_rank;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	nnd     = Ain->nnd;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;

	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dia */
	for(i=0;i<nnd;i++)
	{
		index[i]   = Ain->index[i];
	}
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_ell(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	g_index = NULL;

	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy ell */
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
			index[j*n+i]   = Ain->index[j*n+i];
			g_index[j*n+i] = Ain->g_index[j*n+i];
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	nprocs  = omp_get_max_threads();

	index   = NULL;
	value   = NULL;

	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy ell */
#pragma omp parallel private(i,j,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr;j++)
	{
		for(i=is;i<ie;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
			index[j*n+i]   = Ain->index[j*n+i];
		}
	}
}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;

	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy ell */
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
			index[j*n+i]   = Ain->index[j*n+i];
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_jds(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	nnz     = Ain->nnz;
	is      = Ain->is;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy jds */
	for(i=0;i<n;i++)
	{
		row[i] = Ain->row[i];
	}
	for(j=0;j<maxnzr+1;j++)
	{
		ptr[j] = Ain->ptr[j];
	}
	for(j=0;j<maxnzr;j++)
	{
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			value[i]   = Ain->value[i];
			index[i]   = Ain->index[i];
			g_index[i] = Ain->g_index[i];
		}
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*col,*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	nnz     = Ain->nnz;
	is      = Ain->is;
	nprocs  = omp_get_max_threads();

	col     = NULL;
	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nprocs*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,col,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy jds */
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	col[my_rank] = Ain->col[my_rank];
	for(j=0;j<Ain->col[my_rank]+1;j++)
	{
		ptr[my_rank*(maxnzr+1) + j] = Ain->ptr[my_rank*(maxnzr+1) + j];
	}
	for(i=is;i<ie;i++)
	{
		row[i] = Ain->row[i];
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{
			value[i]   = Ain->value[i];
			index[i]   = Ain->index[i];
		}
	}
}

	err = lis_matrix_set_jds(nnz,col,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,col,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	maxnzr  = Ain->maxnzr;
	nnz     = Ain->nnz;
	is      = Ain->is;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(4,row,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,row,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,row,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,row,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,row,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy jds */
	for(i=0;i<n;i++)
	{
		row[i] = Ain->row[i];
	}
	for(j=0;j<maxnzr+1;j++)
	{
		ptr[j] = Ain->ptr[j];
	}
	for(j=0;j<maxnzr;j++)
	{
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			value[i]   = Ain->value[i];
			index[i]   = Ain->index[i];
		}
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,row,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_bsr(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsr */
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
			g_index[bj] = Ain->g_index[bj];
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsr */
#pragma omp parallel private(i,j,k,bi,bj)
{
#pragma omp for
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
		}
	}
}
	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsr */
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_bsc(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;

	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsc */
	for(i=0;i<nc+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
	for(bi=0;bi<nc;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
			g_index[bj] = Ain->g_index[bj];
		}
	}

	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(4,bptr,bindex,value,g_index);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSC);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsr */
#pragma omp parallel private(i,j,k,bi,bj)
{
#pragma omp for
	for(i=0;i<nc+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
#pragma omp for
	for(bi=0;bi<nc;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
		}
	}
}
	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSC);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	is      = Ain->is;
	nnz     = bnnz*bnr*bnc;
	bs      = bnr*bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy bsr */
	for(i=0;i<nc+1;i++)
	{
		bptr[i] = Ain->bptr[i];
	}
	for(bi=0;bi<nc;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(k=0;k<bs;k++)
			{
				value[bj*bs+k]   = Ain->value[bj*bs+k];
			}
			bindex[bj]  = Ain->bindex[bj];
		}
	}

	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSC);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_vbr(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*row,*col,*ptr,*bptr,*bindex,*g_index,*g_col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;
	nnz     = Ain->nnz;

	row     = NULL;
	col     = NULL;
	ptr     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;
	g_col   = NULL;

	row = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_col = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( g_col==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy vbr */
	for(i=0;i<nc+1;i++)
	{
		col[i]   = Ain->col[i];
		g_col[i] = Ain->g_col[i];
	}
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
		row[i]  = Ain->row[i];
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			k    = Ain->ptr[bj];
			for(j=Ain->col[Ain->bindex[bj]];j<Ain->col[Ain->bindex[bj]+1];j++)
			{
				for(i=Ain->row[bi];i<Ain->row[bi+1];i++)
				{
					value[k]   = Ain->value[k];
					k++;
				}
			}
			bindex[bj]  = Ain->bindex[bj];
			g_index[bj] = Ain->g_index[bj];
			ptr[bj+1]   = Ain->ptr[bj+1];
		}
	}
	ptr[0] = Ain->ptr[0];

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(8,row,col,ptr,bptr,bindex,value,g_index,g_col);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	(*Aout)->g_col   = g_col;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*row,*col,*ptr,*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;
	nnz     = Ain->nnz;

	row     = NULL;
	col     = NULL;
	ptr     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	row = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy vbr */
#pragma omp parallel private(i,j,k,bi,bj)
{
#pragma omp for
	for(i=0;i<nc+1;i++)
	{
		col[i]   = Ain->col[i];
	}
#pragma omp for
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
		row[i]  = Ain->row[i];
	}
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			k    = Ain->ptr[bj];
			for(j=Ain->col[Ain->bindex[bj]];j<Ain->col[Ain->bindex[bj]+1];j++)
			{
				for(i=Ain->row[bi];i<Ain->row[bi+1];i++)
				{
					value[k]   = Ain->value[k];
					k++;
				}
			}
			bindex[bj]  = Ain->bindex[bj];
			ptr[bj+1]   = Ain->ptr[bj+1];
		}
	}
}
	ptr[0] = Ain->ptr[0];

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,bnnz,is,nprocs,my_rank;
	int			bi,bj,bs,nr,nc,bnr,bnc;
	int			*row,*col,*ptr,*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	bnnz    = Ain->bnnz;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;
	nnz     = Ain->nnz;

	row     = NULL;
	col     = NULL;
	ptr     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;

	row = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy vbr */
	for(i=0;i<nc+1;i++)
	{
		col[i]   = Ain->col[i];
	}
	for(i=0;i<nr+1;i++)
	{
		bptr[i] = Ain->bptr[i];
		row[i]  = Ain->row[i];
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			k    = Ain->ptr[bj];
			for(j=Ain->col[Ain->bindex[bj]];j<Ain->col[Ain->bindex[bj]+1];j++)
			{
				for(i=Ain->row[bi];i<Ain->row[bi+1];i++)
				{
					value[k]   = Ain->value[k];
					k++;
				}
			}
			bindex[bj]  = Ain->bindex[bj];
			ptr[bj+1]   = Ain->ptr[bj+1];
		}
	}
	ptr[0] = Ain->ptr[0];

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,row,col,ptr,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_dns(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	int			*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;

	value   = NULL;
	g_index = NULL;

	g_index = (int *)lis_malloc( gn*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(2,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dns */
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
		}
		g_index[j] = Ain->g_index[j];
	}

	err = lis_matrix_set_dns(value,*Aout);
	if( err )
	{
		lis_free2(2,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;

	value   = NULL;

	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free(value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dns */
#pragma omp parallel private(i,j)
{
	for(j=0;j<gn;j++)
	{
#pragma omp for
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
		}
	}
}

	err = lis_matrix_set_dns(value,*Aout);
	if( err )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,maxnzr,is,ie,nprocs,my_rank;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;

	value   = NULL;

	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free(value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy dns */
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i]   = Ain->value[j*n+i];
		}
	}

	err = lis_matrix_set_dns(value,*Aout);
	if( err )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_copy_coo(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*row,*col,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	row     = NULL;
	col     = NULL;
	value   = NULL;
	g_index = NULL;

	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy coo */
	for(i=0;i<nnz;i++)
	{
		row[i]     = Ain->row[i];
		col[i]     = Ain->col[i];
		value[i]   = Ain->value[i];
		g_index[i] = Ain->g_index[i];
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*Aout);
	if( err )
	{
		lis_free2(4,row,col,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_COO);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*row,*col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	row     = NULL;
	col     = NULL;
	value   = NULL;

	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy coo */
	for(i=0;i<nnz;i++)
	{
		row[i]     = Ain->row[i];
		col[i]     = Ain->col[i];
		value[i]   = Ain->value[i];
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_COO);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*row,*col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	row     = NULL;
	col     = NULL;
	value   = NULL;

	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy coo */
	for(i=0;i<nnz;i++)
	{
		row[i]     = Ain->row[i];
		col[i]     = Ain->col[i];
		value[i]   = Ain->value[i];
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_COO);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}
