/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "lislib.h"
#include "lis_precon.h"
#include "lis_solver.h"

#define LIS_SOLVERS_LEN			12
#define LIS_PRECON_TYPE_LEN		9
#define NWORK_CG				8
#define NWORK_BICG				8
#define NWORK_CGS				8
#define NWORK_BICGSTAB			8
#define NWORK_BICGSTABL			4
#define NWORK_GPBICG			15
#define NWORK_ORTHOMIN			3
#define NWORK_GMRES				4
#define NWORK_QMR				11
#define NWORK_JACOBI			3
#define NWORK_GS				3
#define NWORK_SOR				3
extern int lis_nwork[];


int lis_precon_create_iter(LIS_PRECON precon)
{
	int	i,n;
	int restart,maxiter;
	int solvers;
	int *options;
	int worklen;
	LIS_SCALAR *params;
	LIS_MATRIX A;

	options                      = precon->options;
	params                       = precon->params;
	params[LIS_PARAMS_RESID]     = params[LIS_PARAMS_PRESID];
	params[LIS_PARAMS_W]         = params[LIS_PARAMS_PW];
	options[LIS_OPTIONS_MAXITER] = options[LIS_OPTIONS_PMAXITER];
	options[LIS_OPTIONS_ELL]     = options[LIS_OPTIONS_PELL];
	options[LIS_OPTIONS_RESTART] = options[LIS_OPTIONS_PRESTART];
	options[LIS_OPTIONS_OUTPUT]  = 0;
	options[LIS_OPTIONS_SOLVER]  = options[LIS_OPTIONS_PSOLVER];
	options[LIS_OPTIONS_PRECON]  = LIS_PRECON_TYPE_NONE;

	restart = options[LIS_OPTIONS_RESTART];
	solvers = options[LIS_OPTIONS_SOLVER];
	maxiter = options[LIS_OPTIONS_MAXITER];

	if( solvers < 1 || solvers > LIS_SOLVERS_LEN )
	{
		SETERR("Parameter LIS_OPTIONS_PSOLVER is %d (Set between 1 to %d)\n",solvers, LIS_SOLVERS_LEN);
		return LIS_FAILS;
	}
	if( maxiter<0 )
	{
		SETERR("Parameter LIS_OPTIONS_MAXITER(=%d) is less than 0\n",maxiter);
		return LIS_FAILS;
	}
	if( (solvers==LIS_SOLVER_GMRES || solvers==LIS_SOLVER_ORTHOMIN) && restart<0 )
	{
		SETERR("Parameter LIS_OPTIONS_PRESTART(=%d) is less than 0\n",restart);
		return LIS_FAILS;
	}
	else if( solvers==LIS_SOLVER_BICGSTABL && restart<1 )
	{
		SETERR("Parameter LIS_OPTIONS_PELL(=%d) is less than 1\n",restart);
		return LIS_FAILS;
	}

	A       = precon->Amat;
	n       = A->n;


	if( A->matrix_type >= LIS_MATRIX_BSR && A->matrix_type <= LIS_MATRIX_VBR)
	{
		options[LIS_OPTIONS_PADDING] = _max(A->bnr,A->bnc) + _max(A->bnr,A->bnc)%2;
	}
	else
	{
		options[LIS_OPTIONS_PADDING] = 0;
	}
	worklen = lis_nwork[solvers];
	if( solvers==LIS_SOLVER_BICGSTABL ) worklen += 2*(restart+1);
	if( solvers==LIS_SOLVER_GMRES     ) worklen +=   (restart+2);
	if( solvers==LIS_SOLVER_ORTHOMIN  ) worklen += 3*(restart+1);
	
	precon->work = (LIS_VECTOR *)lis_malloc( worklen*sizeof(LIS_VECTOR) );
	for(i=0;i<worklen;i++)
	{
		lis_vector_duplicateM(A,&precon->work[i]);		
	}
	precon->worklen = worklen;

	return LIS_SUCCESS;
}

void lis_psolve_iter(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	int n;
	int err;
	int solvers;
	int *options;
	LIS_SCALAR *params,*status;
	LIS_VECTOR *work;
	LIS_MATRIX AA,At;

	/*
	 *  Mx = b
	 *  M  = A
	 */

	n       = precon->Amat->n;
	AA      = precon->Amat;
	At      = precon->At;
	options = precon->options;
	params  = precon->params;
	status  = precon->status;
	work    = precon->work;
	solvers = options[LIS_OPTIONS_SOLVER];

//	lis_precon_create(AA, b, LIS_PRECON_TYPE_NONE, params, options, &precon_i);

	lis_vector_set_all(0.0,x);
	switch( solvers )
	{
	case LIS_SOLVER_CG:
		err = lis_cg(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_BICG:
		err = lis_bicg(b,x,work,params,options,AA,At,NULL,status);
		break;
	case LIS_SOLVER_CGS:
		err = lis_cgs(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_BICGSTAB:
		err = lis_bicgstab(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_BICGSTABL:
		err = lis_bicgstabl(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_GPBICG:
		err = lis_gpbicg(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_ORTHOMIN:
		err = lis_orthomin(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_GMRES:
		err = lis_gmres(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_QMR:
		err = lis_qmr(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_JACOBI:
		err = lis_jacobi(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_GS:
		err = lis_gs(b,x,work,params,options,AA,NULL,status);
		break;
	case LIS_SOLVER_SOR:
		err = lis_sor(b,x,work,params,options,AA,NULL,status);
		break;
	}
//	lis_psolve_iter2(precon->Amat,b,x,precon->params,precon->options,precon->status);
}

void lis_psolvet_iter(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	int n;
	int err;
	int solvers;
	int *options;
	LIS_SCALAR *params,*status;
	LIS_VECTOR *work;
	LIS_MATRIX AA,At;

	/*
	 *  M'x = b
	 *  M   = A
	 */

	n       = precon->Amat->n;
	AA      = precon->Amat;
	At      = precon->At;
	options = precon->options;
	params  = precon->params;
	status  = precon->status;
	work    = precon->work;
	solvers = options[LIS_OPTIONS_SOLVER];


	lis_vector_set_all(0.0,x);
	switch( solvers )
	{
	case LIS_SOLVER_CG:
		err = lis_cg(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_BICG:
		err = lis_bicg(b,x,work,params,options,At,AA,NULL,status);
		break;
	case LIS_SOLVER_CGS:
		err = lis_cgs(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_BICGSTAB:
		err = lis_bicgstab(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_BICGSTABL:
		err = lis_bicgstabl(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_GPBICG:
		err = lis_gpbicg(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_ORTHOMIN:
		err = lis_orthomin(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_GMRES:
		err = lis_gmres(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_QMR:
		err = lis_qmr(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_JACOBI:
		err = lis_jacobi(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_GS:
		err = lis_gs(b,x,work,params,options,At,NULL,status);
		break;
	case LIS_SOLVER_SOR:
		err = lis_sor(b,x,work,params,options,At,NULL,status);
		break;
	}
//	lis_psolve_iter2(precon->At,b,x,precon->params,precon->options,precon->status);
}
